#!/usr/bin/env julia

using Printf
using Random

function usage()
    prog = basename(Base.source_path())
#    println(STDERR,"Usage ",prog," <rows> <cols> <y1> <y2> <x1> <x2> <no. of threads><lamda> <no. of iter>")
    println(stderr,"Usage ",prog," <rows> <cols> <y1> <y2> <x1> <x2> <lamda> <no. of iter>")
    println(stderr,"\t<rows>   - number of rows")
    println(stderr, "\t<cols>    - number of cols\n")
    println(stderr, "\t<y1> 	 - y1 value of the speckle\n")
    println(stderr, "\t<y2>      - y2 value of the speckle\n")
    println(stderr, "\t<x1>       - x1 value of the speckle\n")
    println(stderr, "\t<x2>       - x2 value of the speckle\n")
#    println(STDERR, "\t<no. of threads>  - no. of threads\n")
    println(stderr, "\t<lamda>   - lambda (0,1)\n")
    println(stderr, "\t<no. of iter>   - number of iterations\n")
    exit(1)
end


function main(args)
    if length(args) != 8
      usage()
    end
    rows = parse(Int64,args[1])
    cols = parse(Int64,args[2])
    if ((rows % 16) != 0) | ((cols % 16) != 0)
        println(stderr,"rows and cols must be mutiples of 16")
        exit(1)
    end
    r1 = parse(Int32,args[3])
    r2 = parse(Int32,args[4])
    c1 = parse(Int32,args[5])
    c2 = parse(Int32,args[6])
    lambda = parse(Float32,args[7])
    niter = parse(Int32,args[8])

    size_R = (r2 - r1 + 1) * (c2 - c1 + 1);

    iN::Array{Int32,1} = [i-1 for i in 1:rows]
    iS::Array{Int32,1} = [i+1 for i in 1:rows]
    jW::Array{Int32,1} = [j-1 for j in 1:cols]
    jE::Array{Int32,1} = [j+1 for j in 1:cols]
    iN[1] = 1
    iS[rows] = rows
    jW[1] = 1
    jE[cols] = cols

    dN::Matrix{Float32} = zeros(cols,rows)
    dS::Matrix{Float32} = zeros(cols,rows)
    dW::Matrix{Float32} = zeros(cols,rows)
    dE::Matrix{Float32} = zeros(cols,rows)

    println("Randomizing the input matrix")

    @ccall srand48(7::Clong)::Cvoid

    I::Matrix{Float32} = zeros(cols,rows)
    for j in 1:size(I, 1)
        for i in 1:size(I, 2)
            I[i, j] = (@ccall drand48()::Cdouble)
        end
    end

    J = exp.(I)
    c::Matrix{Float32} = zeros(cols,rows)

    println("Start the SRAD main loop")

    sum::Float32 = 0
    sum2::Float32 = 0
    for iter in 1:niter
        sum = 0
        sum2 = 0
        for j in r1:r2
            for i in c1:c2
                tmp = J[i + 1,j + 1]
                sum += tmp
                sum2 += tmp * tmp
            end
        end

        meanROI::Float32 = sum / size_R
        varROI::Float32 = (sum2 / size_R) - meanROI * meanROI
        q0sqr::Float32 = varROI / (meanROI * meanROI)

        Threads.@threads for j in 1:size(J,2)
            for i in 1:size(J,1)
                Jc = J[i,j]

                # directional derivates
                dN[i,j] = J[i,iN[j]] - Jc
                dS[i,j] = J[i,iS[j]] - Jc
                dW[i,j] = J[jW[i],j] - Jc
                dE[i,j] = J[jE[i],j] - Jc

                G2 = (dN[i,j] * dN[i,j] + dS[i,j] * dS[i,j] + dW[i,j] * dW[i,j] +
                    dE[i,j] * dE[i,j]) /
                    (Jc * Jc)

                L = (dN[i,j] + dS[i,j] + dW[i,j] + dE[i,j]) / Jc

                num = (0.5 * G2) - ((1.0 / 16.0) * (L * L))
                den = 1 + (0.25 * L)
                qsqr = num / (den * den)

                # diffusion coefficent (equ 33)
                den = (qsqr - q0sqr) / (q0sqr * (1 + q0sqr))
                c[i,j] = 1.0 / (1.0 + den)

                # saturate diffusion coefficent
                if (c[i,j] < 0)
                    c[i,j] = 0
                elseif (c[i,j] > 1)
                    c[i,j] = 1
                end
            end
        end

        Threads.@threads for j in 1:size(J,2)
            for i in 1:size(J,1)
                # diffusion coefficient
                cN = c[i,j]
                cS = c[i,iS[j]]
                cW = c[i,j]
                cE = c[jE[i],j]

                # divergence (equ 58)
                D = cN * dN[i,j] + cS * dS[i,j] + cW * dW[i,j] + cE * dE[i,j]

                # image update (equ 61)
                J[i,j] += 0.25 * lambda * D
            end
        end
    end

    for j in 1:size(J,2)
        for i in 1:size(J,1)
            @printf "%.5f " J[i, j]
        end
        println()
    end

    println("Computation Done")
end

main(ARGS)
